// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include <stdio.h>

#include "stdafx.h"

#include "board.h"
#include "io.h"
#include "distance.h"
#include "util.h"

board initial_board;
board inner;


void copy_board(board from, board to)
{
	int i, j;

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
		to[i][j] = from[i][j];
}

void save_initial_board(board b)
{
	copy_board(b, initial_board);

	get_sokoban_position(b, &global_initial_player_y, &global_initial_player_x);
	global_boxes_in_level = boxes_in_level(b);
}


void get_sokoban_position(board b, int *y, int *x)
{
	int i,j;

	for (i = 0; i < height ; i++)
	for (j = 0; j < width; j++)
	{
		if (b[i][j] & SOKOBAN)
		{
			*y = i;
			*x = j;
			return;
		}
	}
	print_board(b);
	exit_with_error("Sokoban not found\n");
}

void init_inner(board b)
{
	board c;
	int i, j;

	clear_boxes(b, c);
	expand_sokoban_cloud(c);

	// set inner

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
	{
		inner[i][j] = 0;
		if (c[i][j] & SOKOBAN)
			inner[i][j] = 1;
	}
}


static int y_x_to_index_table[MAX_SIZE][MAX_SIZE];
static int index_to_x[MAX_INNER];
static int index_to_y[MAX_INNER];

int index_num;


int y_x_to_index(int y, int x)
{
	int val = y_x_to_index_table[y][x];

	if (val < 0)
		exit_with_error("y_x_to_index");

	if (val >= index_num)
	{
		exit_with_error("val out of range");
	}

	return val;
}

void index_to_y_x(int index, int *y, int *x)
{
	if ((index < 0) || (index >= index_num))
	{
		printf("index: %d index_num: %d\n", index, index_num);
		exit_with_error("index_error");
	}

	*x = index_to_x[index];
	*y = index_to_y[index];
}

void init_index_x_y()
{
	int i, j;
	index_num = 0;

	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			y_x_to_index_table[i][j] = -1;

			if (inner[i][j] == 0)
				continue;

			y_x_to_index_table[i][j] = index_num;
			index_to_x[index_num] = j;
			index_to_y[index_num] = i;
			index_num++;

			if (index_num >= MAX_INNER)
				exit_with_error("inner too big");

		}
	}
}

void clear_boxes(board b, board board_without_boxes)
{
	int i, j;

	copy_board(b, board_without_boxes);

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
		board_without_boxes[i][j] &= ~BOX;
}


void clear_sokoban_inplace(board b)
{
	int i, j;

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
		b[i][j] &= ~SOKOBAN;
}

int board_contains_boxes(board b)
{
	int i, j;

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
	if (b[i][j] & BOX)
		return 1;
	return 0;
}


void turn_boxes_into_walls(board b)
{
	int i, j;

	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			if (b[i][j] & BOX)
				b[i][j] = WALL;
		}
	}
}



void expand_sokoban_cloud(board b)
{
	int queue_y[MAX_SIZE*MAX_SIZE];
	int queue_x[MAX_SIZE*MAX_SIZE];

	int i, j, y, x, next_y, next_x;
	int queue_len = 0;
	int queue_pos = 0;

	// init queue from existing player positions
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (b[i][j] & SOKOBAN)
			{
				queue_y[queue_len] = i;
				queue_x[queue_len] = j;
				queue_len++;
			}

	while (queue_pos < queue_len)
	{
		y = queue_y[queue_pos];
		x = queue_x[queue_pos];

		for (i = 0; i < 4; i++)
		{
			next_y = y + delta_y[i];
			next_x = x + delta_x[i];

			if (b[next_y][next_x] & SOKOBAN) continue;
			if (b[next_y][next_x] & OCCUPIED) continue;

			b[next_y][next_x] |= SOKOBAN;
			queue_y[queue_len] = next_y;
			queue_x[queue_len] = next_x;
			queue_len++;
		}
		queue_pos++;
	}
}


UINT_64 get_board_hash(board b)
{
	UINT_64 hash = 1;
	int i, j;

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
		hash = hash * 12345 + b[i][j];

	return hash;
}

void board_to_bytes(board b, UINT_8 *data)
{
	int i, j;
	int pos = 0;

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
		data[pos++] = b[i][j];
}


void bytes_to_board(UINT_8 *data, board b)
{
	int i, j;
	int pos = 0;

	for (i = 0; i < height; i++)
	for (j = 0; j < width; j++)
		b[i][j] = data[pos++];
}

int all_boxes_in_place(board b)
{
	int i, j;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if ((b[i][j] & BOX) == 0) continue;
		
			if ((b[i][j] & TARGET) == 0) 
				return 0; // not on target
		}
	return 1;
}

int board_is_solved(board b)
{
	return all_boxes_in_place(b);
}


void print_board(board b)
{
	show_board(b);
}


int boxes_in_level(board b)
{
	int i, j;
	int sum = 0;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (b[i][j] & BOX)
				sum++;
	return sum;
}

void zero_board(board b)
{
	int i, j;
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			b[i][j] = 0;
}


void clear_targets_inplace(board b)
{
	int i, j;
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			b[i][j] &= ~TARGET;
}


void keep_boxes_in_inner(board b)
{
	int i, j;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if (inner[i][j] == 0)
				if (b[i][j] & BOX)
					b[i][j] = WALL;
		}
}



int get_indices_of_boxes(board b, int *indices)
{
	int i, y, x;
	int n = 0;

	for (i = 0; i < index_num; i++)
	{
		index_to_y_x(i, &y, &x);
		if (b[y][x] & BOX)
			indices[n++] = i;
	}
	return n;
}

int get_indices_of_targets(board b, int *indices)
{
	int i, y, x;
	int n = 0;

	for (i = 0; i < index_num; i++)
	{
		index_to_y_x(i, &y, &x);
		if (b[y][x] & TARGET)
			indices[n++] = i;
	}
	return n;
}

int get_indices_of_bases(board b, int *indices)
{
	int i, y, x;
	int n = 0;

	for (i = 0; i < index_num; i++)
	{
		index_to_y_x(i, &y, &x);
		if (b[y][x] & BASE)
			indices[n++] = i;
	}
	return n;
}

void clear_bases_inplace(board b)
{
	int i, j;
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			b[i][j] &= ~BASE;
}