// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdio.h>

#include "bfs.h"
#include "corral.h"
#include "util.h"
#include "deadlock.h"

int debug_corral = 0;

#define MAX_CORRALS_NUM 70


int corral_candidates[MAX_CORRALS_NUM];
int corral_candidates_num;

int corral_component_used[MAX_CORRALS_NUM];
int corral_sokoban_comp;
int corral_comps_num;

int_board corral_d;
board corral;


void get_pusher_and_dest(int i, int j, int k,
	int *pusher_y, int *pusher_x, int *dest_y, int *dest_x)
{
	*pusher_y = i + delta_y[k];
	*pusher_x = j + delta_x[k];

	*dest_y = i - delta_y[k];
	*dest_x = j - delta_x[k];
}

int get_direction_from_pusher_dest(int pusher_y, int pusher_x, int dest_y, int dest_x,
	int *box_y, int *box_x)
{
	int i;
	for (i = 0; i < 4; i++)
	{
		if ((pusher_y + delta_y[i] * 2) == dest_y)
			if ((pusher_x + delta_x[i] * 2) == dest_x)
			{
				*box_y = pusher_y + delta_y[i];
				*box_x = pusher_x + delta_x[i];
				return i;
			}
	}
	exit_with_error("no direction");
	return -1;
}


void get_corral_boxes(board orig_b, board corral_boxes)
{
	// get all the boxes that touch the global "corral"
	// also mark the corral area as player.
	int i, j, k;

	clear_boxes(orig_b, corral_boxes);
	clear_sokoban_inplace(corral_boxes);

	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			if (corral[i][j] == 0) continue;

			corral_boxes[i][j] |= SOKOBAN;

			for (k = 0; k < 8; k++)
				if (orig_b[i + delta_y_8[k]][j + delta_x_8[k]] & BOX)
					corral_boxes[i + delta_y_8[k]][j + delta_x_8[k]] |= BOX;
		}
	}
}

int is_possible_push_into_corral(board b,
	int pusher_y, int pusher_x,
	int dest_y, int dest_x)
{
	int box_y, box_x, direction;

	if (b[pusher_y][pusher_x] == WALL) return 0;
	if (b[dest_y][dest_x] & OCCUPIED) return 0;

	// only look for pushes into the corral
	if (corral[pusher_y][pusher_x]) return 0;
	if (corral[dest_y][dest_x] == 0) return 0;

	direction = get_direction_from_pusher_dest(pusher_y, pusher_x, dest_y, dest_x, &box_y, &box_x);

	if (deadlock_in_direction(b, box_y, box_x, direction)) 
		return 0;

	return 1;
}

int is_possible_push_aside_corral(board b,
	int pusher_y, int pusher_x,
	int dest_y, int dest_x)
{
	int box_y, box_x, direction;

	// can't push if occupied
	if (b[pusher_y][pusher_x] & OCCUPIED) return 0;
	if (b[dest_y][dest_x] & OCCUPIED) return 0;

	if (corral[pusher_y][pusher_x]) return 0;
	if (corral[dest_y][dest_x]) return 0;

	direction = get_direction_from_pusher_dest(pusher_y, pusher_x, dest_y, dest_x, &box_y, &box_x);

	if (deadlock_in_direction(b, box_y, box_x, direction))
		return 0;

	return 1;
}



int is_corral_active(board b_in) 
// still work to do inside corral
{
	int i, j;
	board corral_boxes;
	
	get_corral_boxes(b_in, corral_boxes);
	// gets the boxes around the corral, AND marks inner corral with sokoabn

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if (corral[i][j])
				if (corral_boxes[i][j] == TARGET)
					exit_with_error("debug corral");

			if (corral[i][j])
				if (corral_boxes[i][j] == (TARGET | SOKOBAN)) // a target that is not occupied
					return 1;

			if (corral_boxes[i][j] == BOX) return 1; // a box not on a target
		}
	return 0;
}



void set_corral_options(board b_in, board corral_options)
// mark the boxes that can be pushed
{
	int i, j, k;
	int pusher_y, pusher_x, dest_y, dest_x;
	board corral_boxes;

	get_corral_boxes(b_in, corral_boxes);

	clear_boxes(b_in, corral_options);

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if ((corral_boxes[i][j] & BOX) == 0)
				continue;

			for (k = 0; k < 4; k++)
			{
				get_pusher_and_dest(i, j, k, &pusher_y, &pusher_x, &dest_y, &dest_x);

				if (is_possible_push_into_corral(corral_boxes, pusher_y, pusher_x, dest_y, dest_x))
					if (b_in[pusher_y][pusher_x] & SOKOBAN)
						corral_options[i][j] |= BOX;
			}
		}	
}


void get_corral_candidates(board b)
{
	int y, x;
	int i, j, k;
	int other;

	// find areas that are touched by the player.

	corral_comps_num = mark_connectivities(b, corral_d);

	if (corral_comps_num >= MAX_CORRALS_NUM)
		return;

	get_sokoban_position(b, &y, &x);

	corral_sokoban_comp = corral_d[y][x];
	if (debug_corral)
		printf("corral_sokoban_comp= %d\n", corral_sokoban_comp);

	for (i = 0; i < corral_comps_num; i++)
		corral_candidates[i] = 0;

	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			if ((b[i][j] & BOX) == 0) continue;

			for (k = 0; k < 4; k++)
			{
				if (corral_d[i + delta_y[k]][j + delta_x[k]] == corral_sokoban_comp)
				{
					other = corral_d[i - delta_y[k]][j - delta_x[k]];
					if ((other != 1000000) && (other != corral_sokoban_comp))
						corral_candidates[other] = 1;
				}
			}

		}
	}


	corral_candidates_num = 0;
	for (i = 0; i < corral_comps_num; i++)
		corral_candidates_num += corral_candidates[i];

	if (debug_corral)
	{
		printf("corral candidates: ");
		for (i = 0; i < corral_comps_num; i++)
			if (corral_candidates[i])
				printf("%d ", i);
		printf("\n");
	}
}


int init_corral_with_comp(int comp1, int comp2, board b_in)
{
	int i, j;

	zero_board(corral);

	for (i = 0; i < corral_comps_num; i++)
		corral_component_used[i] = 0;

	corral_component_used[comp1] = 1;
	corral_component_used[comp2] = 1;
	
	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			if ((corral_d[i][j] == comp1) || (corral_d[i][j] == comp2))
			{
				corral[i][j] = 1;
			}
		}
	}

	return is_corral_active(b_in);
}

void update_corral_with_comp(int comp1)
{
	int i, j;

	corral_component_used[comp1] = 1;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (corral_d[i][j] == comp1)
				corral[i][j] = 1;

}


int box_touches_corral_and_non_corral(board b, int y, int x)
{
	int i;

	int touches_corral = 0;
	int touches_non_corral = 0;

	for (i = 0; i < 4; i++)
	{
		if (corral[y + delta_y[i]][x + delta_x[i]]) touches_corral = 1;

		if ((b[y + delta_y[i]][x + delta_x[i]] & OCCUPIED) == 0)
			if (corral[y + delta_y[i]][x + delta_x[i]] == 0) 
				touches_non_corral = 1;
	}

	return touches_corral & touches_non_corral;

}

int should_add_component_to_corral(board corral_boxes, int *component)
{
	int i, j, k, d;
	int pusher_x, pusher_y;
	int dest_x, dest_y;
	int can_push_inside, can_push_aside;

	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			if ((corral_boxes[i][j] & BOX) == 0)
				continue;

			if (box_touches_corral_and_non_corral(corral_boxes, i, j) == 0)
				continue;

			for (k = 0; k < 4; k++)
			{
				get_pusher_and_dest(i, j, k, &pusher_y, &pusher_x, &dest_y, &dest_x);

				can_push_inside = is_possible_push_into_corral(corral_boxes, pusher_y, pusher_x, dest_y, dest_x);
				if (corral_boxes[pusher_y][pusher_x] & OCCUPIED) can_push_inside = 0;

				can_push_aside = is_possible_push_aside_corral(corral_boxes, pusher_y, pusher_x, dest_y, dest_x);
					
				if ((can_push_inside == 0) && (can_push_aside == 0))
					continue;

				// by previous check, we know that pusher is outside the corral

				if (can_push_inside)
				{
					d = corral_d[pusher_y][pusher_x];

					if (d != 1000000) // could be a box in front of the corral
					if ((d != corral_sokoban_comp) && (corral_component_used[d] == 0))
					{
						*component = d;
						return 1;
					}
				}

				if (can_push_aside)
				{
					d = corral_d[dest_y][dest_x];

					if (d == 1000000) exit_with_error("this is strange");
					if ((d != corral_sokoban_comp) && (corral_component_used[d] == 0))
					{
						*component = d;
						return 1;
					}
				}
			} // k
		}   // j
	}  // i
	return 0;
} 
	


int can_push_box_out_of_corral(board b)
{
	int i, j, k;
	int pusher_x, pusher_y;
	int dest_x, dest_y;
	int box_y, box_x, direction;

	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			if ((b[i][j] & BOX) == 0)
				continue;

			if (inner[i][j] == 0) continue;

			for (k = 0; k < 4; k++)
			{
				get_pusher_and_dest(i, j, k, &pusher_y, &pusher_x, &dest_y, &dest_x);
				direction = get_direction_from_pusher_dest(pusher_y, pusher_x, dest_y, dest_x, &box_y, &box_x);

				// can't push if occupied
				if (b[pusher_y][pusher_x] & OCCUPIED) continue;
				if (b[dest_y][dest_x] & OCCUPIED) continue;

				if (deadlock_in_direction(b, box_y, box_x, direction)) continue;

				if (corral[pusher_y][pusher_x] == 0)
					if (corral[dest_y][dest_x] == 0)
					{
						if (debug_corral)
							printf("can push aside box %d %d\n\n", i, j);
						return 1;
					}
			}
		}
	}

	return 0;
}



int check_pi_corral(board corral_boxes, board b_in)
{
	int i, j, k;
	int pusher_x, pusher_y;
	int dest_x, dest_y;

	// verify that all the boxes on the corral boundary can actually be pushed inside
	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			if ((corral_boxes[i][j] & BOX) == 0) continue;

			for (k = 0; k < 4; k++)
			{
				get_pusher_and_dest(i, j, k, &pusher_y, &pusher_x, &dest_y, &dest_x);

				if (is_possible_push_into_corral(corral_boxes, pusher_y, pusher_x, dest_y, dest_x) == 0)
					continue;

				if (corral_boxes[pusher_y][pusher_x] & OCCUPIED) continue; // (i,j) is not on the corral boundary

				if (corral[dest_y][dest_x] == 0) exit_with_error("not right1");
				if (corral[pusher_y][pusher_x] == 1) exit_with_error("not right2");

				if ((b_in[pusher_y][pusher_x] & SOKOBAN) == 0)
				{
					if (debug_corral)
					{
						print_board(corral);
						printf("box in front of corral: %d %d\n", i, j);
					}
					return 0;
				}
			}
		}
	}

	return 1;
}

int expand_corral(board b_in)
{
	// given the current corral, try to expand it by adding components, until it can
	// be proven to be a pi-corral.
	// Or report that it's not a pi-corral once the expansions stop.
	board corral_boxes;
	int res, d;

	do
	{
		get_corral_boxes(b_in, corral_boxes);

		if (debug_corral)
		{
			printf("corral boxes:\n");
			print_board(corral_boxes);
			my_getch();
		}

		res = should_add_component_to_corral(corral_boxes, &d);
		if (res)
		{
			if (debug_corral)
				printf("adding component %d\n", d);
			update_corral_with_comp(d);
		}
	} while (res);

	if (can_push_box_out_of_corral(corral_boxes))
		return 0;

	if (check_pi_corral(corral_boxes, b_in) == 0)
		return 0;

	if (debug_corral)
		printf("pi corral found!\n");
	return 1;
}


int touches_two_comps[MAX_CORRALS_NUM][MAX_CORRALS_NUM];

void init_touches_two_comps(board b)
{
	int i, j, k, l;
	int d1, d2;

	if (corral_comps_num >= MAX_CORRALS_NUM)
	{
		print_board(b);
		printf("corrals: %d\n", corral_comps_num);
		exit_with_error("Too many corrals");
	}

	for (i = 0; i < MAX_CORRALS_NUM; i++)
		for (j = 0; j < MAX_CORRALS_NUM; j++)
			touches_two_comps[i][j] = 0;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if ((b[i][j] & BOX) == 0) continue;

			for (k = 0; k < 4; k++)
				for (l = 0; l < 4; l++)
				{
					d1 = corral_d[i + delta_y[k]][j + delta_x[k]];
					d2 = corral_d[i + delta_y[l]][j + delta_x[l]];

					if ((d1 != 1000000) && (d2 != 1000000))
					{
						touches_two_comps[d1][d2] = 1;
					}
				}
		}
}


int detect_corral(board b, board corral_options)
{
	int i,j,res;

	if (debug_corral)
	{
		printf("detecting corral in board:\n");
		print_board(b);
	}

	get_corral_candidates(b);

	if (corral_comps_num >= MAX_CORRALS_NUM)
		return 0; // do not handle corrals when there are too many of them

	for (i = 0; i < corral_comps_num; i++)
	{
		if (corral_candidates[i] == 0) continue;

		if (debug_corral)
			printf("trying component %d\n", i);
		res = init_corral_with_comp(i, i, b);
		if (res == 0) continue;

		res = expand_corral(b);
		if (res == 1)
		{
			set_corral_options(b, corral_options);
			return 1;
		}
	}

	init_touches_two_comps(b);

	// look for two-rooms corral
	for (i = 0; i < corral_comps_num; i++)
	{

		if (i == corral_sokoban_comp) continue;

		for (j = i + 1; j < corral_comps_num; j++)
		{
			if (j == corral_sokoban_comp) continue;

			if (corral_candidates[j] == 0) 
				if (corral_candidates[i] == 0) continue;

			if ((touches_two_comps[i][j]) == 0) continue;

			if (debug_corral)
				printf("trying components %d %d\n", i, j);
			res = init_corral_with_comp(i, j, b);
			if (res == 0) continue;

			res = expand_corral(b);
			if (res == 1)
			{
				set_corral_options(b, corral_options);
				return 1;
			}
		}
	}
	
	return 0;
}