// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "curry.h"
#include "board.h"
#include "moves.h"
#include "episode.h"
#include "util.h"
#include "optimize.h"
#include "sol.h"
#include "match_deadlock.h"


void setup_board_with_k_boxes_inner(board b, int k)
{
	int target_indices[MAX_INNER];
	int box_indices[MAX_INNER];
	int chosen[MAX_INNER];
	int n, i, j, y, x;

	n = get_indices_of_boxes(initial_board, box_indices);
	n = get_indices_of_targets(initial_board, target_indices);
	
	copy_board(initial_board, b);
	
	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
			if (b[i][j] & BOX)
				b[i][j] |= BASE;
			

	// remove some boxes
	get_k_of_n(k, n, chosen);
	for (i = 0; i < n; i++)
		if (chosen[i] == 0)
		{
			index_to_y_x(box_indices[i], &y, &x);
			b[y][x] &= ~BOX;
			b[y][x] &= ~BASE;
		}

	// remove some targets
	get_k_of_n(k, n, chosen);
	for (i = 0; i < n; i++)
		if (chosen[i] == 0)
		{
			index_to_y_x(target_indices[i], &y, &x);
			b[y][x] &= ~TARGET;
		}

	expand_sokoban_cloud(b);
}

void setup_board_with_k_boxes(board b, int k)
{
	int res;

	do
	{
		setup_board_with_k_boxes_inner(b, k);

		res = check_matching_deadlock(b); 

		if (board_is_solved(b)) 
			res = 1;

	} while (res == 1);
}


void get_weights_of_moves(board b, move *moves, int moves_num, float *p, episodes *es)
{

	int from, to, i, j;
	move m;
	float pseudocount;
	move suggested_moves[1000];
	int same_conf[1000];
	episode *current_episode;
	int value;

	static float weight_matrix[MAX_INNER][MAX_INNER];

	int b_has_box[MAX_INNER];
	int b_has_target[MAX_INNER];

	if (boxes_in_level(b) == 1) // no relevant past episodes
	{
		for (i = 0; i < moves_num; i++)
			p[i] = 1;
		return;
	}

	for (i = 0; i < index_num; i++)
		for (j = 0; j < index_num; j++)
			weight_matrix[i][j] = (float)(0.000001);


	for (i = 0; i < index_num; i++)
		b_has_box[i] = b_has_target[i] = 0;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if (b[i][j] & BOX)
				b_has_box[y_x_to_index(i, j)] = 1;
			if (b[i][j] & TARGET)
				b_has_target[y_x_to_index(i, j)] = 1;
		}


	pseudocount = 1 / (float)(index_num * 10);

	for (i = 0; i < es->episodes_num; i++)
	{
		current_episode = es->e + i;

		get_suggestion_from_episode(b_has_box, b_has_target,
			current_episode, &m, &value);

		suggested_moves[i] = m;
		to = m.to;

		if (m.to < 0) exit_with_error("negative move");

		same_conf[i] = current_episode->has_target[to] ^ b_has_target[to] ^ 1;
	
	}

	for (i = 0; i < es->episodes_num; i++)
	{
		m = suggested_moves[i];

		if (same_conf[i] == 0) continue;

		from = m.from;
		to = m.to;

		weight_matrix[from][to] += 1;
		
		for (j = 0; j < index_num; j++)
		{
			weight_matrix[from][j] += pseudocount;
			weight_matrix[j][to]   += pseudocount;
		}
		
	}


	for (i = 0; i < moves_num; i++)
	{
		from = moves[i].from;
		to = moves[i].to;

		p[i] = weight_matrix[from][to];
	}
}


int curry_time_exceeded()
{
	if (int(time(0) - start_time) > time_limit)
	{
		strcpy(global_fail_reason, "Time limit exceeded");
		return 1;
	}
	return 0;
}


double get_board_value(board b, episodes *es)
{
	int i, j;
	episode *current_episode;
	int value;
	move m;
	double total_value = 0;

	int b_has_box[MAX_INNER];
	int b_has_target[MAX_INNER];

	if (boxes_in_level(b) == 1) return 1; 

	for (i = 0; i < index_num; i++)
		b_has_box[i] = b_has_target[i]  = 0;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if (b[i][j] & BOX)
				b_has_box[y_x_to_index(i, j)] = 1;
			if (b[i][j] & TARGET)
				b_has_target[y_x_to_index(i, j)] = 1;
		}

	for (i = 0; i < es->episodes_num; i++)
	{
		current_episode = es->e + i;

		get_suggestion_from_episode(b_has_box, b_has_target,
			current_episode, &m, &value);

		total_value += value;
	}

	if (es->episodes_num == 0)
		return 0;

	return total_value / es->episodes_num;
}


int get_player_position_from_tree(expansion_data *e)
{
	board b;
	int y, x;

	bytes_to_board(e->b, b);
	get_sokoban_position(b, &y, &x);
	return y_x_to_index(y, x);
}

void get_episode_from_tree(tree *t, expansion_data *e, episode *sol)
{
	board b;
	int i, j;
	sol->len = e->depth;

	sol->player_at_end = get_player_position_from_tree(e);

	while (e->depth > 0)
	{
		sol->moves[e->depth - 1] = get_move_to_node(t, e);

		e = e->father;
	}

	if (e->depth != 0) exit_with_error("bad root");

	bytes_to_board(e->b, b);
	get_indices_of_bases(b, sol->bases);
	get_indices_of_targets(b, sol->targets);
	sol->box_num = boxes_in_level(b);

	for (i = 0; i < index_num; i++)
		sol->has_base[i] = sol->has_target[i] = 0;

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if (b[i][j] & BASE)
				sol->has_base[y_x_to_index(i, j)] = 1;
			if (b[i][j] & TARGET)
				sol->has_target[y_x_to_index(i, j)] = 1;
		}

	optimize_episode(sol);

	if (verbose >= 5)
	{
		printf("\n\nAdded episode:\n");
		print_episode(sol);
		printf("\n\n");
	}

}
