// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <string.h>
#include <stdio.h>

#include "moves.h"
#include "deadlock.h"
#include "distance.h"
#include "util.h"

int debug_deadlock = 0;

board forbidden_tunnel;

void set_forbidden_tunnel()
{
	int i, j, k, n;
	board b;
	int base;

	zero_board(forbidden_tunnel);

	clear_boxes(initial_board, b);
	clear_sokoban_inplace(b);

	// look for
	// ######
	//
	// ######
	for (i = 1; i < (height-1); i++)
	{
		for (j = 0; j < width; j++)
		{
			for (k = 0; k < width; k++)
			{
				if (b[i - 1][j + k] != WALL) break;
				if (b[  i  ][j + k] != SPACE) break;
				if (b[i + 1][j + k] != WALL) break;
			}
			if (k >= 2)
			{
				n = k;

				base = -1;
				// look for base
				for (k = 0; k < n; k++)
					if (initial_board[i][j + k] & BOX)
						break;

				if (k < n)
					base = k;

				for (k = 0; k < n; k++)
				{
					forbidden_tunnel[i][j + k] = 1;
					b[i][j + k] |= SOKOBAN;
				}

				if (base == -1)
					forbidden_tunnel[i][j] = 0;
				else
					forbidden_tunnel[i][j + base] = 0;
			}
		}
	}
	
	// look for
	// # #
	// # #
	// # #
	for (i = 0; i < height; i++)
	{
		for (j = 1; j < (width - 1); j++)
		{
			for (k = 0; k < height; k++)
			{
				if (b[i + k][j - 1] != WALL) break;
				if (b[i + k][j] != SPACE) break;
				if (b[i + k][j + 1] != WALL) break;
			}
			if (k >= 2)
			{
				n = k;

				base = -1;
				for (k = 0; k < n; k++)
					if (initial_board[i + k][j] & BOX)
						break;

				if (k < n)
					base = k;

				for (k = 0; k < n; k++)
				{
					forbidden_tunnel[i + k][j] = 1;
					b[i + k][j] |= SOKOBAN;
				}

				if (base == -1)
					forbidden_tunnel[i][j] = 0;
				else
					forbidden_tunnel[i + base][j] = 0;
			}
		}
	}
}


int is_2x2_deadlock_at_place(board b, int i, int j)
{
	int y, x;

	for (y = 0; y < 2; y++)
		for (x = 0; x < 2; x++)
			if ((b[i + y][j + x] & OCCUPIED) == 0)
				return 0;

	for (y = 0; y < 2; y++)
		for (x = 0; x < 2; x++)
			if (b[i + y][j + x] & BOX)
				if ((b[i + y][j + x] & TARGET) == 0)
					return 1;
	return 0;
}


void mark_deadlock_2x2(board c_inp, board d)
{
	int i, j,y,x;
	int sum, stuck;
	board c;
	unsigned char val;

	copy_board(c_inp, c);
	clear_sokoban_inplace(c);

	zero_board(d);

	for (i = 0; i < (height - 1); i++)
	for (j = 0; j < (width - 1); j++)
	{

		// look for patterns composed of 3 boxes/walls such as
		// $?
		// #$

		sum = 0;
		stuck = 0; // boxes that will deadlock if they become fixed
		for (y = 0; y < 2; y++)
			for (x = 0; x < 2; x++)
			{
				val = c[i + y][j + x];
				if (val & OCCUPIED) sum++;
				if ((val != WALL) && ((val & TARGET) == 0)) stuck++;
				// it could be a box or an empty square, but immobilizing a box there would be a mistake
			}

		if (sum == 3) 
			if (stuck)
		{
			for (y = 0; y < 2; y++)
			for (x = 0; x < 2; x++)
				d[i + y][j + x] = 1;
		}
		
		// look for pattern
		// #? 
		// ?#

		if ((c[i][j] == WALL) && (c[i + 1][j + 1] == WALL))
		{
			if (c[i][j + 1] == 0)
				d[i][j + 1] = 1;
			if (c[i + 1][j] == 0)
				d[i + 1][j] = 1;
		}

		// look for pattern
		// ?# 
		// #?

		if ((c[i][j+1] == WALL) && (c[i + 1][j] == WALL))
		{
			if (c[i][j] == 0)
				d[i][j] = 1;
			if (c[i + 1][j + 1] == 0)
				d[i + 1][j + 1] = 1;
		}
		
	}
}

int deadlock_in_direction(board b, int y, int x, int d)
{
	// check if a freeze deadlock will occur after pushing the box in direction d.

	//
	//   00
	//  3  1
	//  3 $1
	//   22
	static int square_place_y[4][2] =
	{
		{ -2, -2 }, { -1, 0 }, { 1, 1 }, { -1, 0 }
	};

	static int square_place_x[4][2] =
	{
		{ -1, 0 }, { 1, 1 }, { -1, 0 }, { -2, -2 }
	};

	int dy, dx, new_y, new_x;
	int dy2, dx2; // push sideway
	int res_0, res_1;
	int wall1 = 0, wall2 = 0, wall3 = 0, wall4 = 0;

	dy = delta_y[d];
	dx = delta_x[d];
	dy2 = delta_y[(d + 1) & 3]; // sideway
	dx2 = delta_x[(d + 1) & 3];
	new_y = y + dy;
	new_x = x + dx;

	if (inner[new_y][new_x] == 0) return 1; // safeguard before calling y_x_to_index
	if (impossible_place[y_x_to_index(new_y, new_x)]) 
		return 1;

	if ((b[y][x] & BOX) == 0) exit_with_error("missing box");
	if (b[new_y][new_x] & OCCUPIED) exit_with_error("no space");

	// first, look for two walls:
	//   #
	//  #$
	//   ^ = (y,x)


	if ( 
		((b[new_y][new_x] & TARGET) == 0) && 
		(b[new_y + dy][new_x + dx] == WALL)
		)
	{
		if (b[new_y + dy2][new_x + dx2] == WALL) return 1;
		if (b[new_y - dy2][new_x - dx2] == WALL) return 1;
	}

	// check patterns of type
	//   1  #$   2
	//   3   $#	 4
	//       ^

	if (b[new_y + dy][new_x + dx] & BOX)
	{
		if (b[new_y + dy2][new_x + dx2] == WALL) wall4 = 1;
		if (b[new_y - dy2][new_x - dx2] == WALL) wall3 = 1;
		if (b[new_y + dy + dy2][new_x + dx + dx2] == WALL) wall2 = 1;
		if (b[new_y + dy - dy2][new_x + dx - dx2] == WALL) wall1 = 1;

		if (((wall1 + wall4) == 2) || ((wall2 + wall3) == 2))
		{
			if ((b[new_y][new_x]           & TARGET) == 0) return 1;
			if ((b[new_y + dy][new_x + dx] & TARGET) == 0) return 1;
		}
	}

	// temporarily push the box and check 2x2 deadlocks
	b[y][x] &= ~BOX;
	b[new_y][new_x] |= BOX;

	res_0 = is_2x2_deadlock_at_place(b, y + square_place_y[d][0], x + square_place_x[d][0]);
	res_1 = is_2x2_deadlock_at_place(b, y + square_place_y[d][1], x + square_place_x[d][1]);

	// restore box
	b[new_y][new_x] &= ~BOX;
	b[y][x] |= BOX;

	if (res_0 || res_1) return 1;

	return 0;
}

int is_freeze_deadlock(board b, int y, int x)
{
	// check patterns of type
	//    #O     #$    
	//     $#     *#

	int i;
	int next_y, next_x, dy, dx;

	if ((b[y][x] & BOX) == 0) exit_with_error("check caller");

	for (i = 0; i < 4; i++)
	{
		next_y = y + delta_y[i];
		next_x = x + delta_x[i];

		dy = delta_y[(i + 1) & 3];
		dx = delta_x[(i + 1) & 3];

		if ((b[y][x] & TARGET) == 0)
		{
			if (b[next_y][next_x] & OCCUPIED)
			{
				if ((b[y + dy][x + dx] == WALL) && (b[next_y - dy][next_x - dx] == WALL))
					return 1;
				if ((b[y - dy][x - dx] == WALL) && (b[next_y + dy][next_x + dx] == WALL))
					return 1;
			}
		}

		if (b[y][x] & TARGET)
		{
			if (b[next_y][next_x] == BOX) // not on target
			{
				if ((b[y + dy][x + dx] == WALL) && (b[next_y - dy][next_x - dx] == WALL))
					return 1;
				if ((b[y - dy][x - dx] == WALL) && (b[next_y + dy][next_x + dx] == WALL))
					return 1;
			}
		}

	}

	return 0;

}



int is_inefficient_position(board b, int box_index)
{
	int box_y, box_x;
	int val[8];
	int i, j;

	// check positions like
	//  #
	// $ 
	// ##

	index_to_y_x(box_index, &box_y, &box_x);

	// never prune moves that push a box to a target
	if (b[box_y][box_x] & TARGET)
		return 0;

	if (forbidden_tunnel[box_y][box_x])
		return 1;


	for (i = 0; i < 8; i++)
		val[i] = b[box_y + delta_y_8[i]][box_x + delta_x_8[i]];


	for (j = 0; j < 8; j+=2)
	{
		// check positions like
		//     
		//  -$# 
		//  # #
		//
		// last move must have been up/down, both are silly

		if ( val[(j + 2) & 7] == WALL)
		if ( val[(j + 3) & 7] == WALL)
		if ( val[(j + 5) & 7] == WALL)
		if ( val[(j + 6) & 7] != WALL)
		if ((val[(j + 6) & 7] & SOKOBAN) == 0)
		{
			return 1;
		}

		if ( val[(j + 3) & 7] == WALL)
		if ( val[(j + 5) & 7] == WALL)
		if ( val[(j + 6) & 7] == WALL)
		if ( val[(j + 2) & 7] != WALL)
		if ((val[(j + 2) & 7] & SOKOBAN) == 0)
		{
			return 1;
		}

		// check positions like
		//   @ 
		//  -$# 
		//  # #
		//
		// could just as well park it
		
		if (val[(j + 2) & 7] == WALL)
		if (val[(j + 3) & 7] == WALL)
		if (val[(j + 5) & 7] == WALL)
		if (val[(j + 6) & 7] != WALL)
		if (val[(j + 0) & 7] & SOKOBAN)
		{
			return 1;
		}

		if (val[(j + 3) & 7] == WALL)
		if (val[(j + 5) & 7] == WALL)
		if (val[(j + 6) & 7] == WALL)
		if (val[(j + 2) & 7] != WALL)
		if (val[(j + 0) & 7] & SOKOBAN)
		{
			return 1;
		}
		
		// check
		// ???
		// -$$
		// #?#
		if (val[(j + 3) & 7] == WALL)
		if (val[(j + 5) & 7] == WALL)
		if (val[(j + 2) & 7] & BOX)
			if ((val[(j + 6) & 7] & SOKOBAN) == 0)
			{
				return 1;
			}

		if (val[(j + 3) & 7] == WALL)
		if (val[(j + 5) & 7] == WALL)
		if (val[(j + 6) & 7] & BOX)
			if ((val[(j + 2) & 7] & SOKOBAN) == 0)
			{
				return 1;
			}
	}

	return 0;

}


void check_deadlock_in_moves(board b_in, move *moves, int moves_num, int *valid)
{
	int i, dest, dest_x,dest_y;
	board b,c;

	copy_board(b_in, b);

	for (i = 0; i < moves_num; i++)
	{
		if (debug_deadlock)
		{
			printf("checking move: ");
			print_move(moves + i);
		}

		valid[i] = 1;

		dest = moves[i].to;
		index_to_y_x(dest, &dest_y, &dest_x);


		if (impossible_place[dest])
		{
			valid[i] = 0;
			if (debug_deadlock)
				printf("rejected: impossible place\n");
			continue;
		}

		copy_board(b, c);		
		apply_move(c, moves + i);


		if (is_inefficient_position(c, moves[i].to))
		{
			valid[i] = 0;
			if (debug_deadlock)
				printf("rejected on inefficient move\n");

			continue;
		}

		if (is_freeze_deadlock(c, dest_y, dest_x))
		{
			valid[i] = 0;
			if (debug_deadlock)
				printf("freeze deadlock\n");
			continue;
		}


		if (debug_deadlock)
			printf("ok\n");
	}

}

