// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdlib.h>

#include "episode.h"
#include "select.h"
#include "util.h"
#include "engine.h"
#include "optimize.h"
#include "curry.h"
#include "sol.h"
#include "leaf.h"
#include "challenges.h"

int seed = 12345;
int subcases_num = 80;

tree search_tree;
episodes past_episodes;

int walk_id;


void init_search_tree()
{
	init_tree(&search_tree, 22);
}


void allocate_past_episodes(int n)
{
	allocate_episodes(&past_episodes, n);
	past_episodes.episodes_num = n;
}


expansion_data *restart_walk(tree *t)
{
	expansion_data *e;

	e = t->expansions;
	
	if (e->node->deadlocked) return 0;

	walk_id++;

	return e;
}

expansion_data *select_node_to_expand(tree *t, int *son)
{
	expansion_data *e;
	int found_leaf = 0;

	e = restart_walk(t);
	e->walk_id = walk_id;

	do 
	{		
		if (e->node->deadlocked) exit_with_error("got to deadlocked");

		*son = select_son(t, e);

		if (*son == -1) break;

		if (son_is_expanded(t, e, *son) == 0)
		{
			found_leaf = 1;
			break;
		}

		e = go_to_son(t, e, *son);
		e->walk_id = walk_id;

	} 
	while (1);
	
	if (found_leaf)
		return e;

	return 0;
}


void do_rollout(tree *t, expansion_data *e, int son)
{
	expansion_data *expanded_node;
	char info;

	double best_val;
	int steps_since_best;

	best_val = e->value;
	steps_since_best = 0;

	do
	{
		if (e->depth >= (MAX_SOL_LEN - 2)) return;
		if (e->node->deadlocked) exit_with_error("got to deadlocked");

		expanded_node = e;

		if (verbose >= 5)
		{
			printf("expanding node:\n");
			print_node(t, expanded_node, 0);
			printf("with move %d\n", son);
		}
			
		expand_node(t, expanded_node, son);

		e = last_expansion(t);
		e->walk_id = walk_id;

		if (verbose >= 5)
		{
			printf("node after expansion:\n");
			print_node(t, e, 0);
		}

		if (node_is_solved(t, e)) return;
		if (tree_nearly_full(t, &info)) return;
		if (e->node->deadlocked) return;

		son = select_son(t, e);

		if ((son == -1) || (son_is_expanded(t, e, son)))
			return;

		// check if we still make progress
		if (e->value < best_val)
		{
			best_val = e->value;
			steps_since_best = 0;
		}
		else
			steps_since_best++;

	} 
	while (steps_since_best < 10);
}


int solve_board(board b, tree *t, int max_expansions, episode *sol, char *status)
{
	int son;
	int solved = 0;
	expansion_data *e, *root;
	char info;

	walk_id = 0;

	reset_tree(t);

	set_root(t, b);
	root = t->expansions;

	e = restart_walk(t);
	if (e) e->walk_id = walk_id;

	while (1)
	{
		if (curry_time_exceeded()) break;

		if (root->node->deadlocked)
		{
			*status = '-';
			return -1;
		}

		if (t->expansions_num >= max_expansions)
		{
			*status = 'A';
			break;
		}

		e = select_node_to_expand(t, &son);

		if (e == 0)
		{ // use the alternative way 
			e = pick_node_using_weight(t, &son);
			if (e == 0)
			{
				*status = '-';
				return -1;
			}
		}

		if (e->node->deadlocked) exit_with_error("got to deadlocked");

		do_rollout(t, e, son);

		e = last_expansion(t);
		if (node_is_solved(t, e))
		{
			solved = 1;
			break;
		}

		if (tree_nearly_full(t, &info))
		{
			*status = info;
			break;
		}
	}

	sol->difficulty = t->expansions_num;

	if (solved)
	{
		*status = '.';
		get_episode_from_tree(t, e, sol);
	}

	return solved;
}


void update_past_episodes(episode *e)
{
	// replace one past episode with the new solution

	int place = find_easiest_episode(&past_episodes);
	past_episodes.e[place] = *e;
}



int solve_one_challenge(int k, int i, char *status)
{
	int search_size = 20000;
	board b;
	int solved;
	episode e;

	static int counter = 0;

	inv_epsilon = 2*k + 1;

	get_challenge(k, i, b);

	solved = solve_board(b, &search_tree, search_size, &e, status);

	update_board_difficulty(k, i, e.difficulty);

	if (solved == -1) // impossible board - replace the challenge
	{
		replace_challenge(k, i);
		return -1;
	}

	if (solved == 1)
	{
		mark_board_as_solved(k, i);

		e.seq = counter++;
		e.difficulty = get_board_difficulty(k, i);

		update_past_episodes(&e);
	}

	return solved;
}

int solve_instances_with_k_boxes(int k)
{
	// returns 1 if at least one level was solved

	int i, solved;
	int solved_level = 0;
	int episodes_num = past_episodes.episodes_num;
	int failed_num = 0;
	char status;

	if (k == 0) return 0;
	if (solved_all_challenges(k)) return 0;

	printf("%2d ", k);

	for (i = 0; i < episodes_num; i++)
	{
		if (curry_time_exceeded()) 
			break;

		if (is_board_solved(k , i))
		{
			printf(".");
			continue;
		}

		solved = solve_one_challenge(k, i, &status);

		printf("%c", status);

		if (solved != 1) failed_num++;
		if (solved == 1) solved_level = 1;

		if ((solved == 1) && (k == global_boxes_in_level))
			return 1;
	}

	printf("  %2d", failed_num);
	printf("\n");

	return solved_level;
}



void solve_k_control(int k)
{
	int stuck_counter = 0;
	int profficient, res;

	do 
	{
		if (curry_time_exceeded()) return;

		res = solve_instances_with_k_boxes(k);

		if (res == 0)
			stuck_counter++;
		else
			stuck_counter = 0;

		profficient = solved_all_challenges(k);

		if (solution_index(&past_episodes) != -1) break; // found a solution to the level
		if (k == global_boxes_in_level)           break; // don't try k boxes more than once

		if (stuck_counter >= 6)
			break;

	} while (profficient == 0);
}


void init_dummy_past_episodes()
{
	int i;
	for (i = 0; i < past_episodes.episodes_num; i++)
	{
		past_episodes.e[i].box_num = 0;
		past_episodes.e[i].len = 0;
		past_episodes.e[i].seq = -1;
	}
}

void solve_all_k(int box_num)
{
	int k;

	for (k = 1; k <= box_num; k++)
	{
		if (curry_time_exceeded()) return;
		solve_k_control(k);
	}
}


int curry_engine(board b)
{
	int  box_num;
	int solved;

	srand(seed);	

	box_num = boxes_in_level(b);
	printf("%d boxes\n", box_num);

	while (curry_time_exceeded() == 0)
	{
		init_challenges(box_num, subcases_num);
		init_dummy_past_episodes(); // episodes with zero boxes

		solve_all_k(box_num);

		free_challenges(box_num);

		if (solution_index(&past_episodes) != -1) break; // has a solution
	}
	
	solved = save_solution_if_found(&past_episodes);

	return solved;
}

