// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include "graph.h"
#include "bfs.h"
#include "moves.h"
#include "util.h"

int vertices_num;

typedef struct vertex
{
	int shift[4];
	int push;
	int weight;
	int active;
	int src;
} vertex;

vertex vertices[MAX_INNER * 4];

// vertex index is: (position of box) | (player position)
// player position is 2 bits.
// shift,push is in the index to which we move by applying this operation.


void init_active(board b)
{
	int i, j;
	int y, x, base_y, base_x;

	// "active" tests if the player can be positioned in this side of the box

	for (i = 0; i < index_num; i++)
		for (j = 0; j < 4; j++)
			vertices[i * 4 + j].active = 0;

	for (i = 0; i < index_num; i++)
	{
		index_to_y_x(i, &base_y, &base_x);

		if (b[base_y][base_x] & WALL)
		{
			// this can happen when boxes are converted to walls
			// in order to find possible moves of a single box
			for (j = 0; j < 4; j++)
				vertices[i * 4 + j].active = 0;
			continue;
		}


		for (j = 0; j < 4; j++)
		{
			x = base_x + delta_x[j];
			y = base_y + delta_y[j];

			if (b[y][x] & WALL)
				vertices[i * 4 + j].active = 0;
			else
			{
				vertices[i * 4 + j].active = 1;
			}
		}
	}
}


void init_shift(board b)
{
	int i, j;
	int y, x;
	int_board zones;
	int from, to;

	for (i = 0; i < vertices_num; i++)
		for (j = 0; j < 4; j++)
			vertices[i].shift[j] = 0;

	for (i = 0; i < index_num; i++)
	{
		index_to_y_x(i, &y, &x);

		if (b[y][x] & OCCUPIED)
			continue;

		b[y][x] |= BOX;
		mark_connectivities_around_place(b, zones, y, x);

		b[y][x] &= ~BOX;

		for (from = 0; from < 4; from++)
		{
			if (vertices[i * 4 + from].active == 0) continue;
			if (zones[y + delta_y[from]][x + delta_x[from]] == 1000000) continue;

			for (to = 0; to < 4; to++)
			{
				if (to == from) continue;
				if (vertices[i * 4 + to].active == 0) continue;
				if (zones[y + delta_y[to]][x + delta_x[to]] == 1000000) continue;

				if (zones[y + delta_y[from]][x + delta_x[from]] ==
					zones[y + delta_y[to]][x + delta_x[to]])
					vertices[i * 4 + from].shift[to] = i * 4 + to;
			}
		}
	} // index of box
}



void init_push(board b)
{
	int i;
	int base_x, base_y;
	int push_x, push_y;
	int new_index;

	for (i = 0; i < vertices_num; i++)
	{
		if (vertices[i].active == 0)
			continue;

		index_to_y_x(i / 4, &base_y, &base_x);

		if (b[base_y][base_x] & OCCUPIED)
		{
			exit_with_error("something is bad");
		}

		push_x = base_x - delta_x[i % 4];
		push_y = base_y - delta_y[i % 4];

		if (b[push_y][push_x] & OCCUPIED)
			vertices[i].push = 0;
		else
		{
			new_index = y_x_to_index(push_y, push_x);

			vertices[i].push = new_index * 4 + (i % 4);
		}
	}
}




void set_graph_weights_to_infinity()
{
	int i;

	for (i = 0; i < vertices_num; i++)
	{
		vertices[i].weight = 1000000;
		vertices[i].src = -1;
	}
}



void build_graph(board b)
{
	// set up the active vertices and edges, don't compute anything yet

	vertices_num = index_num * 4;

	if (board_contains_boxes(b))
		exit_with_error("boxes in board");
	
	init_active(b);
	init_shift(b);

	init_push(b);

	set_graph_weights_to_infinity();
}

void clear_weight_around_cell(int index)
{
	int j;

	for (j = 0; j < 4; j++)
	{
		if (vertices[index * 4 + j].active)
			vertices[index * 4 + j].weight = 0;
	}
}



void do_graph_iterations()
{
#define MAX_BFS_NODES (MAX_INNER * 4)
	static int current[MAX_BFS_NODES];
	static int next[MAX_BFS_NODES];
	static int visited[MAX_BFS_NODES];
	int i, j, next_vertex;
	int current_num = 0;
	int next_num = 0;
	int d = 0;
	int v;

	for (i = 0; i < vertices_num; i++)
		visited[i] = 0;

	// init from start vertices
	for (i = 0; i < vertices_num; i++)
	{
		if (vertices[i].weight == 0)
		{
			vertices[i].src = -1;
			next[next_num++] = i;
			visited[i] = 1;
		}
	}

	while (next_num)
	{
		d++;

		for (i = 0; i < next_num; i++)
			current[i] = next[i];
		current_num = next_num;
		next_num = 0;

		for (i = 0; i < current_num; i++)
		{
			v = current[i];

			for (j = 0; j < 4; j++)
			{
				next_vertex = vertices[v].shift[j];
				if (next_vertex == 0) continue;

				if (vertices[next_vertex].active == 0)
					exit_with_error("internal error1");

				if (visited[next_vertex]) continue;

				visited[next_vertex] = 1;
				vertices[next_vertex].weight = vertices[v].weight;
				vertices[next_vertex].src = v;
				next[next_num++] = next_vertex;
			}

			next_vertex = vertices[v].push;

			if (next_vertex == 0) continue;

			if (vertices[next_vertex].active == 0)
				exit_with_error("internal error2");


			if (visited[next_vertex]) continue;
			visited[next_vertex] = 1;
			vertices[next_vertex].weight = vertices[v].weight + 1;
			vertices[next_vertex].src = v;
			next[next_num++] = next_vertex;

			if (next_num >= MAX_BFS_NODES)
				exit_with_error("bfs overflow");
		} // loop on current vertices with weight d
	} // while next d

}



int get_weight_around_cell(int index)
{
	int j;
	int min = 1000000;

	for (j = 0; j < 4; j++)
	{
		if (vertices[index * 4 + j].active == 0) continue;

		if (vertices[index * 4 + j].weight < min)
			min = vertices[index * 4 + j].weight;
	}
	return min;
}


int get_box_moves_from_graph(int start_index, move *moves)
{
	int i, j, src;
	int moves_num = 0;

	for (i = 0; i < index_num; i++)
	{
		if (current_impossible[i]) continue; // can't push box here - probably simple deadlock

		for (j = 0; j < 4; j++)
		{
			if (vertices[i * 4 + j].weight == 1000000)
				continue;

			if (vertices[i * 4 + j].weight == 0)
				continue; // this is one of the start positions - not a push

			src = vertices[i * 4 + j].src;
			if ((src/4) == i)
				continue; // last move was not a push, but a player move

			moves[moves_num].from = start_index;
			moves[moves_num].to = i;
			moves[moves_num].player_position = j;

			moves_num++;

		}
	}
	return moves_num;
}


void clear_shift()
{
	int i, j;
	for (i = 0; i < (index_num * 4); i++)
		for (j = 0; j < 4; j++)
			vertices[i].shift[j] = 0;
}



void update_shift(board b, int box_y, int box_x)
{
	// using the Sokoban cloud, mark which shifts are possible
	int i, from, to, index;
	int has_sokoban[4] = { 0, 0, 0, 0 };

	index = y_x_to_index(box_y, box_x) * 4;

	for (i = 0; i < 4; i++)
		if (b[box_y + delta_y[i]][box_x + delta_x[i]] & SOKOBAN)
			has_sokoban[i] = 1;

	for (from = 0; from < 4; from++)
	{
		if (has_sokoban[from] == 0) continue;

		for (to = 0; to < 4; to++)
			if (has_sokoban[to])
				vertices[index + from].shift[to] = index + to;

	}
}

void find_box_moves_graph(board b_in, int box_index)
{
	board b;
	int i, j;
	int box_y, box_x;
	int next_box_y, next_box_x;

	int sokoban_y, sokoban_x;
	int next_sok_y, next_sok_x;

	vertices_num = index_num * 4;
	int direction;

	int d = 0;
	int index, next_index;
	int current[MAX_SIZE * MAX_SIZE];
	int next[MAX_SIZE * MAX_SIZE];
	int current_size = 0;
	int next_size = 0;
	int index_after_push;

	copy_board(b_in, b);

	clear_shift();

	set_graph_weights_to_infinity();

	index_to_y_x(box_index, &box_y, &box_x);
	if ((b[box_y][box_x] & BOX) == 0)
	{
		print_board(b);
		exit_with_error("missing box3");
	}

	for (i = 0; i < 4; i++)
	{
		sokoban_y = box_y + delta_y[i];
		sokoban_x = box_x + delta_x[i];
		if (b[sokoban_y][sokoban_x] & SOKOBAN)
		{
			vertices[box_index * 4 + i].weight = d;
			current[current_size++] = box_index * 4 + i;
		}
	}

	b[box_y][box_x] &= ~BOX;

	// invariant: queue contains box locations and all nearby player positions.
	// all that is left is to push

	while (current_size)
	{
		d++;
		for (i = 0; i < current_size; i++)
		{
			index = current[i];
			direction = index & 3;
			index_to_y_x(index / 4, &box_y, &box_x);

			sokoban_y = box_y + delta_y[direction];
			sokoban_x = box_x + delta_x[direction];

			
			next_box_y = box_y - delta_y[direction];
			next_box_x = box_x - delta_x[direction];

			next_sok_y = sokoban_y - delta_y[direction];
			next_sok_x = sokoban_x - delta_x[direction];

			if (b[next_box_y][next_box_x] & OCCUPIED) continue;
			
			if (current_impossible_board[next_box_y][next_box_x]) continue;

			next_index = y_x_to_index(next_box_y, next_box_x) * 4 + direction;

			if (vertices[next_index].weight != 1000000) continue;

			// so the box can be pushed there, and this is a new move. now compute surroundings

			vertices[next_index].weight = d;
			vertices[next_index].src = index;
			next[next_size++] = next_index;
			index_after_push = next_index;

			clear_sokoban_inplace(b);
			b[next_box_y][next_box_x] |= BOX;
			b[next_sok_y][next_sok_x] |= SOKOBAN;
			expand_sokoban_cloud(b);

			update_shift(b, next_box_y, next_box_x);

			for (j = 0; j < 4; j++)
			{
				next_sok_y = next_box_y + delta_y[j];
				next_sok_x = next_box_x + delta_x[j];

				if ((b[next_sok_y][next_sok_x] & SOKOBAN) == 0) continue;

				next_index = y_x_to_index(next_box_y, next_box_x) * 4 + j;

				if (vertices[next_index].weight != 1000000) continue;

				vertices[next_index].weight = d;
				vertices[next_index].src = index_after_push;
				next[next_size++] = next_index;
			}

			b[next_box_y][next_box_x] &= ~BOX;
		} // current_size

		for (i = 0; i < next_size; i++)
			current[i] = next[i];
		current_size = next_size;
		next_size = 0;
	} // BFS queue
}
