// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <stdio.h>

#include "hungarian.h"
#include "global.h"
#include "util.h"

static int c[MAX_BOXES][MAX_BOXES];
static int n;

int V[MAX_BOXES];
int U[MAX_BOXES];
int matched_v[MAX_BOXES];
int matched_u[MAX_BOXES];
int match_size;

// for BFS
int src_of_v[MAX_BOXES];
int src_of_u[MAX_BOXES];
int visited_v[MAX_BOXES];
int visited_u[MAX_BOXES];

int v_list[MAX_BOXES];
int v_list_num;
int u_list[MAX_BOXES];
int u_list_num;


void set_hungarian_problem(int size, int mat[MAX_BOXES][MAX_BOXES])
{
	int i, j;

	n = size;
	for (i = 0; i < n; i++)
	for (j = 0; j < n; j++)
		c[i][j] = mat[i][j];
}

void print_cost_matrix()
{
	int i, j;

	for (i = 0; i < n; i++)
	{
		for (j = 0; j < n; j++)
			printf("%3d ", c[i][j]);
		printf("\n");
	}
	printf("\n");
}


void init_U_V()
{
	int min = 0;
	int i, j;

	for (i = 0; i < n; i++)
	{
		min = 10000000;
		for (j = 0; j < n; j++)
		{
			if (c[i][j] < min)
				min = c[i][j];
		}
		V[i] = min;
	}

	for (j = 0; j < n; j++)
	{
		min = 10000000;
		for (i = 0; i < n; i++)
		{
			if ((c[i][j] - V[i]) < min)
				min = (c[i][j] - V[i]);
		}
		U[j] = min;
	}

}



void init_match()
{
	int i,j;
	match_size = 0;

	for (i = 0; i < n; i++)
	{
		matched_v[i] = -1;
		matched_u[i] = -1;
	}

	for (i = 0; i < n; i++)
	{
		for (j = 0; j < n; j++)
		{
			if (matched_u[j] != -1) continue;

			if (c[i][j] == (V[i] + U[j]))
			{
				matched_v[i] = j;
				matched_u[j] = i;
				match_size++;
//				printf("added %d %d\n", i, j);
				break; // j iteration
			}
		}
	}


}


void init_hungarian_bfs()
{
	int i;
	v_list_num = 0;

	for (i = 0; i < n; i++)
	{
		src_of_v[i] = -1;
		src_of_u[i] = -1;
		visited_v[i] = 0;
		visited_u[i] = 0;
	}

	for (i = 0; i < n; i++)
	{
		if (matched_v[i] == -1)
		{
			v_list[v_list_num++] = i;
			visited_v[i] = 1;
		}
	}
}


int bfs_v_side()
{
	int t, i,j;
	int changed = 0;

	u_list_num = 0;

	for (t = 0; t < v_list_num; t++)
	{
		i = v_list[t];

		for (j = 0; j < n; j++)
		{
			if (visited_u[j]) continue;

			if ((V[i] + U[j]) == c[i][j])
			{
				visited_u[j] = 1;
				src_of_u[j] = i;
				u_list[u_list_num++] = j;
				changed = 1;
			}
		}
	}
	return changed;
}

int bfs_u_side()
{
	int i, j, t;
	int changed = 0;

	v_list_num = 0;
	for (t = 0; t < u_list_num; t++)
	{
		j = u_list[t];

		if (matched_u[j] == -1)
			continue;

		i = matched_u[j];

		if (visited_v[i] == 1)
			continue;

		v_list[v_list_num++] = i;
		visited_v[i] = 1;
		src_of_v[i] = j;
		changed = 1;
	}
	return changed;
}

void hungarian_bfs(int side) // v = side 0
{
	int res;

	do
	{
		if (side == 0)
		{
			if (v_list_num == 0) exit_with_error("bad hungarian v side\n");
			res = bfs_v_side();
		}
		else
		{
			if (u_list_num == 0) exit_with_error("bad hungarian u side\n");
			res = bfs_u_side();
		}
		side = 1 - side;
	} 
	while (res);
}


int try_to_increase_match()
{
	int i, j;

	for (j = 0; j < n; j++)
	{
		if (matched_u[j] != -1) continue;
		if (visited_u[j] == 0) continue;

		while (j != -1)
		{
			i = src_of_u[j];
			matched_u[j] = i;
			matched_v[i] = j;
			j = src_of_v[i];
		}
		return 1;
	}
	return 0;
}


void add_new_edges()
{
	int i, j;
	int min = 10000000;
	int delta;

	u_list_num = 0;

	// determine delta
	for (i = 0; i < n; i++)
	{
		if (visited_v[i] == 0) continue;

		for (j = 0; j < n; j++)
		{
			if (visited_u[j] == 1) continue;

			delta = c[i][j] - V[i] - U[j];
			if (delta < min)
				min = delta;
		}
	}

	delta = min;

	// update weights
	for (i = 0; i < n; i++)
	{
		if (visited_v[i])
			V[i] += delta;
		if (visited_u[i])
			U[i] -= delta;
	}

	// add vertices to U using new edges
	for (j = 0; j < n; j++)
	{
		if (visited_u[j]) continue;

		for (i = 0; i < n; i++)
		{
			if (visited_v[i] == 0) continue;

			if ((V[i] + U[j]) == c[i][j])
			{
				visited_u[j] = 1;
				src_of_u[j] = i;
				break;
			}
		}  // i

		if (visited_u[j])
			u_list[u_list_num++] = j;
	} // j
}


void verify_hungarian_invariant()
{
	int i, j;

	for (i = 0; i < n; i++)
	for (j = 0; j < n; j++)
		if ((V[i] + U[j]) > c[i][j])
		{
			exit_with_error("invariant failed\n");
		}
}


void add_edge_to_match()
{
	int res;

	init_hungarian_bfs();

	hungarian_bfs(0); // side = v
	// do several v-u iterations, starting from v

	while (1)
	{
		res = try_to_increase_match();
		if (res) return;

		add_new_edges();

		hungarian_bfs(1); // side = u
	}
}


void verify_hungarian_solution()
{
	int i,j;

	verify_hungarian_invariant();

	// confirm match

	for (i = 0; i < n; i++)
	{
		j = matched_v[i];
		if (j == -1) exit_with_error("not perm");
		if (matched_u[j] != i) exit_with_error("not perm");

		if ((V[i] + U[j]) != c[i][j])
			exit_with_error("match problem");
	}

}

int is_match_full()
{
	int i;

	for (i = 0; i < n; i++)
	if (matched_v[i] == -1)
		return 0;
	return 1;
}

void do_hungarian_iterations()
{
	while (is_match_full() == 0)
	{
		add_edge_to_match();
		verify_hungarian_invariant();
	}

	verify_hungarian_solution();
}


void set_hungarian_solution(hungarian_solution *sol)
{
	int i;
	sol->weight = 0;
	
	for (i = 0; i < n; i++)
	{
		sol->match_v[i] = matched_v[i];
		sol->match_u[i] = matched_u[i];
		sol->U[i] = U[i];
		sol->V[i] = V[i];
		sol->weight += c[i][matched_v[i]];
	}
}




void solve_hungarian(int n, int cost[MAX_BOXES][MAX_BOXES], hungarian_solution *sol)
{
	set_hungarian_problem(n, cost);

	init_U_V();
	init_match();

	do_hungarian_iterations();

	set_hungarian_solution(sol);
}

