// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include <math.h>

#include "leaf.h"
#include "util.h"

// Some routines for picking the expanded leaf when the standard leaf selection fails.
// In general, pick the leaf with the highest probability of reaching it

int best_unexpanded_son(tree *t, expansion_data *e)
{
	int i;
	int best_son = -1;
	float best_weight = 0;
	move_hash_data *mh = t->move_hashes + e->move_hash_place;

	if (e->node->deadlocked) return -1;

	for (i = 0; i < e->moves_num; i++)
	{
		if (son_is_deadlocked(t, e, i)) continue;
		if (son_is_expanded(t, e, i)) continue;

		if ((best_son == -1) || (mh[i].weight > best_weight))
		{
			best_son = i;
			best_weight = mh[i].weight;
		}
	}
	return best_son;
}


double get_weight_of_unexpanded_son(tree *t, expansion_data *e, int son)
{
	int i;
	double total = 0;
	double p;
	double epsilon_prob;
	int n_moves = 0;
	move_hash_data *mh = t->move_hashes + e->move_hash_place;

	extern int inv_epsilon;

	if (e->node->deadlocked) exit_with_error("strange1");
	if (son_is_expanded(t, e, son))	exit_with_error("something is wrong");

	for (i = 0; i < e->moves_num; i++)
	{
		if (son_is_deadlocked(t, e, i)) continue;

		total += mh[i].weight;
		n_moves++;
	}

	if (n_moves == 0) exit_with_error("strange2");

	if (total != 0)
		p = mh[son].weight / total;
	else
		p = 1 / (double)(n_moves);

	epsilon_prob = 1 / (double)(inv_epsilon);

	p = (epsilon_prob / n_moves) + (1 - epsilon_prob)*p;

	if ((p <= 0) || (p > 1))
		exit_with_error("numerical error");

	return -log(p);
}

void set_best_weight(tree *t, expansion_data *e)
{
	int son;
	double weight;

	if (e->node->deadlocked)
	{
		e->best_weight = 1000000;
		return;
	}

	son = best_unexpanded_son(t, e);

	if (son == -1)
	{
		e->best_weight = 1000000;
		return;
	}

	weight = get_weight_of_unexpanded_son(t, e, son);

	if (weight == 1000000) exit_with_error("weight error");

	e->best_weight = e->weight + weight;
}


expansion_data *pick_node_using_weight(tree *t, int *son)
{
	double best_weight = 1000000;
	int i;
	expansion_data *e, *best_node = 0;

	for (i = 0; i < t->expansions_num; i++)
	{
		e = t->expansions + i;

		if (e->best_weight >= best_weight) continue;

		set_best_weight(t, e);

		if (e->best_weight >= best_weight) continue;

		best_weight = e->best_weight;
		best_node = e;
	}

	if (best_node == 0) return 0;

	*son = best_unexpanded_son(t, best_node);
	return best_node;
}

