// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"
#include "moves.h"

#include <stdio.h>

#include "graph.h"

#include "deadlock.h"
#include "distance.h"
#include "corral.h"
#include "fixed_boxes.h"
#include "util.h"
#include "bfs.h"

int current_impossible[MAX_INNER];
board current_impossible_board;


void set_current_impossible(board b_in, int box_y, int box_x)
{
	int i, j, index,y,x;
	board blocked;
	board b;

	copy_board(b_in, b);

	for (i = 0; i < index_num; i++)
	{
		current_impossible[i] = impossible_place[i]; 
		// general impossible places of the level. board edges etc.

		// add boxes that were wallified
		index_to_y_x(i, &y, &x);
		if (b[y][x] == WALL)
			current_impossible[i] = 1;
	}


	b[box_y][box_x] &= ~BOX;

	mark_deadlock_2x2(b, blocked);

	for (i = 0; i < height; i++)
		for (j = 0; j < width; j++)
		{
			if (inner[i][j] == 0) continue;

			if (blocked[i][j])
			{
				index = y_x_to_index(i, j);
				current_impossible[index] = 1;
			}
		}
	

	
	zero_board(current_impossible_board);

	for (i = 0; i < index_num; i++)
		if (current_impossible[i])
		{
			index_to_y_x(i, &y, &x);
			current_impossible_board[y][x] = 1;
		}

}


int find_box_moves(board b, int index, move *moves)
{
	int y, x;
	int moves_num;
	int i;

	index_to_y_x(index, &y, &x);

	for (i = 0; i < 4; i++)
		if (b[y + delta_y[i]][x+delta_x[i]] & SOKOBAN) break;
	if (i == 4) return 0; // player does not touch box

	set_current_impossible(b,y,x); 
	// find situation-specific deadlocks, in addition to the level deadlocks

	find_box_moves_graph(b, index);

	moves_num = get_box_moves_from_graph(index, moves);

	return moves_num;
}

int keep_valid_moves(move *moves, int *valid, int moves_num)
{
	int i;
	int p = 0;

	move local_moves[MAX_MOVES];

	for (i = 0; i < moves_num; i++)
	{
		if (valid[i])
		{
			local_moves[p] = moves[i];
			p++;
		}
	}

	for (i = 0; i < p; i++)
		moves[i] = local_moves[i];

	return p;
}

int find_possible_moves(board b_input, move *moves,	int *has_corral)
{
	int moves_num = 0;
	int i, y, x;
	int box_moves;
	board b,c;
	int valid[MAX_MOVES];
	board corral_options;
	int use_corral = 1;

	*has_corral = 0;

	copy_board(b_input, c);


	clear_bases_inplace(c);

	turn_fixed_boxes_into_walls(c, b);

	if (use_corral)			
		*has_corral = detect_corral(b, corral_options);

	for (i = 0; i < index_num; i++)
	{
		index_to_y_x(i, &y, &x);

		if (*has_corral)
			if ((corral_options[y][x] & BOX) == 0)
				continue;

		if (b[y][x] & BOX)
		{
			box_moves = find_box_moves(b, i , moves + moves_num);
			moves_num += box_moves;

			if (moves_num > (MAX_MOVES - MAX_SIZE*MAX_SIZE))
			{
				if (verbose >= 4)
					printf("too many moves: %d moves\n", moves_num);

				sprintf(global_fail_reason, "Too many moves");
				break; // do not produce other moves
			}
		}
	}


	check_deadlock_in_moves(b, moves, moves_num, valid);

	moves_num = keep_valid_moves(moves, valid, moves_num);

	return moves_num;
}

void apply_move(board b, move *m)
{
	int y, x, p;

	clear_sokoban_inplace(b);

	index_to_y_x(m->from, &y, &x); // box to push: initial position

	if ((b[y][x] & BOX) == 0) exit_with_error("missing box2");
	b[y][x] &= ~BOX;

	index_to_y_x(m->to, &y, &x); // box to push: final position

	if (b[y][x] & OCCUPIED) exit_with_error("occupied place");

	b[y][x] |= BOX;


	p = m->player_position;
	b[y + delta_y[p]][x + delta_x[p]] |= SOKOBAN;

	expand_sokoban_cloud(b);

}


void print_move(move *m)
{
	int move_from_y, move_from_x;
	int move_to_y, move_to_x;

	index_to_y_x(m->from, &move_from_y, &move_from_x);
	index_to_y_x(m->to, &move_to_y, &move_to_x);

	printf("%2d %2d -> %2d %2d (%d) ",
		move_from_y, move_from_x,
		move_to_y, move_to_x, m->player_position);
}




void get_hashes_of_moves(board b, int moves_num, move *moves, UINT_64 *hashes)
{
	board c;
	int i;

	for (i = 0; i < moves_num; i++)
	{
		copy_board(b, c);
		apply_move(c, moves + i);
		hashes[i] = get_board_hash(c);
	}
}

