// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"

#include "episode.h"
#include "util.h"

int optimtimize_verbose = 3;

int player_after_move(move *m)
{
	int y, x;

	index_to_y_x(m->to, &y, &x);
	y += delta_y[m->player_position];
	x += delta_x[m->player_position];
	return y_x_to_index(y, x);
}

void print_episode_moves(episode *e)
{
	int i, n;
	n = e->len;

	move *moves;

	moves = e->moves;

	printf("Episode with %d moves\n", n);
	for (i = 0; i < n; i++)
	{
		print_move(moves + i);
		printf("\n");
	}
}

int is_move_possible(board b, move *m)
{
	move moves[MAX_MOVES];
	int corral, moves_num, i, y, x;
	board c;

	moves_num = find_possible_moves(b, moves, &corral);

	for (i = 0; i < moves_num; i++)
	{
		if (moves[i].from != m->from) continue;
		if (moves[i].to   != m->to) continue;

		// test player's position
		copy_board(b, c);
		apply_move(c, moves + i);

		index_to_y_x(m->to, &y, &x);
		y += delta_y[m->player_position];
		x += delta_x[m->player_position];

		if (c[y][x] & SOKOBAN) return 1;
	}
	return 0;
}

int is_episode_possible(episode *e, int from, int to, int required_player_position)
{
	// check if the episode can be played in moves from (including) until to (excluding)
	// It is the callers responsibility that all other moves are legal !
	// Otherwise the program may crash !

	board b;
	int i,y,x;
	move *moves;

	setup_board_from_episode(b, e);

	moves = e->moves;

	for (i = 0; i < from; i++)
		apply_move(b, moves + i);

	for (i = from; i < to; i++)
	{
		if (is_move_possible(b, moves + i))
			apply_move(b, moves + i);
		else
		{
			if (verbose >= 5)
			{
				printf("detected impossible move:\n");
				print_board(b);
				print_move(moves + i);
				printf("\n");
			}

			return 0;
		}
	}

	if (required_player_position != -1)
	{
		index_to_y_x(required_player_position, &y, &x);
		if ((b[y][x] & SOKOBAN) == 0)
			return 0;
	}

	return 1;
}

void drop_move_from_episode(episode *e, int move_index)
{
	int i,n;

	move *moves;

	moves = e->moves;

	n = e->len;
	for (i = move_index; i < (n - 1); i++)
		moves[i] = moves[i + 1];
	e->len = n - 1;
}


int optimize_episode_at_step_i(episode *e, int i)
{
	episode f;
	int j;

	move *moves;
	move *moves2;

	moves = e->moves;

	for (j = i + 1; j < e->len; j++)
		if (moves[i].to == moves[j].from)
			break;

	if (j == e->len) return 0;

	// a box was moved at step i and then at step j.
	// try to optimize by combining the moves at step i, and combining in step j

	// step i:
	f = *e;
	moves2 = f.moves;
	moves2[i].to = moves[j].to;
	moves2[i].player_position = moves[j].player_position;
	drop_move_from_episode(&f, j);

	if (is_episode_possible(&f, i ,j, player_after_move(moves + j)))
	{
		*e = f;
		return 1;
	}

	// step j:
	f = *e;
	moves2 =  f.moves;
	moves2[j].from = moves[i].from;
	drop_move_from_episode(&f, i);
	if (is_episode_possible(&f, i, j, player_after_move(moves + j)))
	{
		*e = f;
		return 1;
	}

	return 0;
}


int optimize_episode_sweep(episode *e)
{
	int i, res;
	int optimized = 0;

	for (i = 0; i < e->len; i++)
	{
		do
		{
			res = optimize_episode_at_step_i(e, i);
			optimized |= res;
		} 
		while (res == 1);
	}

	return optimized;
}

void optimize_episode(episode *e)
{
	int res;

	episode backup = *e;

	if (optimtimize_verbose >= 5)
		print_episode_moves(e);

	do
	{
		res = optimize_episode_sweep(e);
	} 
	while (res);

	if (is_episode_possible(e, 0, e->len, -1) == 0)
		exit_with_error("wrong optimization");

	if (optimtimize_verbose >= 5)
		print_episode_moves(e);
}

