// Curry Sokoban Solver
// Copyright 2021 Yaron Shoham

#include "stdafx.h"
#include <stdlib.h>

#include "tree.h"
#include "util.h"
#include "engine.h"
#include "leaf.h"
#include "curry.h"

void init_tree(tree *t, int log_max_nodes)
{
	int i;
	unsigned int size; // replace with size_t ?

	// hash_array
	t->hash_size = 1 << (log_max_nodes + 2);
	size = sizeof(int) * t->hash_size;
	if (verbose >= 4)
		printf("Allocating %12d bytes for hash_array\n", size);
	t->hash_array = (int*)malloc(size);
	t->mask = t->hash_size - 1;

	if (t->hash_array == 0) exit_with_error("can't allocate hash_array");
	for (i = 0; i < t->hash_size; i++)
		t->hash_array[i] = -1;

	// nodes
	t->max_nodes = 1 << log_max_nodes;
	size = sizeof(node_element) * t->max_nodes;
	if (verbose >= 4)
		printf("Allocating %12d bytes for %12d nodes\n", size, t->max_nodes);
	t->nodes = (node_element*)malloc(size);
	t->nodes_num = 0;
	if (t->nodes == 0) exit_with_error("can't allocate nodes");

	// expansions
	t->max_expansions = 1 << (log_max_nodes - 6);
	size = sizeof(expansion_data) * t->max_expansions;
	if (verbose >= 4)
		printf("Allocating %12d bytes for %12d expansions\n", size, t->max_expansions);
	t->expansions = (expansion_data *)malloc(size);
	if (t->expansions == 0) exit_with_error("can't allocate expansions list");
	t->expansions_num = 0;

	// boards
	t->boards_size = size * 6;
	if (verbose >= 4)
		printf("Allocating %12d bytes for boards\n", t->boards_size);
	t->boards = (UINT_8 *)malloc(t->boards_size);
	if (t->boards == 0) exit_with_error("can't allocate boards");

	// move-hashes
	t->max_move_hashes = 1 << (log_max_nodes + 1);
	size = sizeof(move_hash_data) * t->max_move_hashes;
	if (verbose >= 4)
		printf("Allocating %12d bytes for %12d move-hashes\n", size, t->max_move_hashes);
	t->move_hashes = (move_hash_data *)malloc(size);
	if (t->move_hashes == 0) exit_with_error("can't allocate move-hashes");
	t->move_hashes_num = 0;

}

void free_tree(tree *t)
{
	free(t->hash_array);
	free(t->nodes);
	free(t->expansions);
	free(t->move_hashes);
	free(t->boards);
}

void reset_tree(tree *t)
{
	int i;

	for (i = 0; i < t->hash_size; i++)
		t->hash_array[i] = -1;
	t->nodes_num = 0;
	t->expansions_num = 0;
	t->move_hashes_num = 0;
	
}

int tree_nearly_full(tree *t, char *reason)
{
	if (t->expansions_num > (t->max_expansions - 2))
	{
		*reason = 'E';
		return 1;
	}
	if (t->nodes_num > (t->max_nodes - MAX_MOVES))
	{
		*reason = 'N';
		return 1;
	}
	if (t->move_hashes_num > (t->max_move_hashes - MAX_MOVES))
	{
		*reason = 'H';
		return 1;
	}

	if (((t->expansions_num + 1) * height * width) >= (t->boards_size))
	{
		*reason = 'B';
		return 1;
	}

	return 0;
}


int find_node_by_hash(tree *t, UINT_64 hash)
{
	int index, place;
	index = hash & t->mask;

	place = t->hash_array[index];
	while (place != -1)
	{
		if (t->nodes[place].hash == hash)
			return place;

		index = (index + 1) & t->mask;
		place = t->hash_array[index];
	}

	return -1;
}

void add_entry_to_hash_array(tree *t, UINT_64 hash, int node_place)
{
	int index,place;
	index = hash & t->mask;

	place = t->hash_array[index];
	while (place != -1)
	{
		if (t->nodes[place].hash == hash) exit_with_error("already in array");
		index = (index + 1) & t->mask;
		place = t->hash_array[index];
	}
	t->hash_array[index] = node_place;
}


void add_node_to_tree(tree *t, UINT_64 hash, int wins)
{
	int place;
	node_element *node;

	place = find_node_by_hash(t, hash);
	if (place != -1) 
		exit_with_error("adding existing hash");

	place = t->nodes_num;
	node = t->nodes + place;
	node->expansion = -1;
	node->hash = hash;
	node->wins = (char)wins;
	node->deadlocked = 0;

	t->nodes_num++;

	if (t->nodes_num >= t->max_nodes)
		exit_with_error("max tree nodes reached");

	add_entry_to_hash_array(t, hash, place);
}


int node_is_solved(tree *t, expansion_data *e)
{
	board b;

	bytes_to_board(e->b, b);

	if (e->depth == 0) return 0;
	// a position can't be solved without making any moves
	// this is for cyclic levels

	return board_is_solved(b);
}


void set_deadlock_status(tree *t, expansion_data *e)
{
	int i,place;
	node_element *node;
	move_hash_data *mh;

	if (e->node->deadlocked)
		return;

	e->node->deadlocked = 0;

	if (node_is_solved(t,e))
		return;
	
	// update the mh[i].deadlock values
	mh = t->move_hashes + e->move_hash_place;
	for (i = 0; i < e->moves_num; i++)
	{
		if (mh[i].deadlocked) continue;

		place = find_node_by_hash(t, mh[i].hash);
		if (place == -1) exit_with_error("missing node");
		node = t->nodes + place;

		if (node->deadlocked)
			mh[i].deadlocked = 1;
	}

	// update node itself

	for (i = 0; i < e->moves_num; i++)
		if (mh[i].deadlocked == 0)
			return;

	e->node->deadlocked = 1;

	if (e->father)
		set_deadlock_status(t, e->father);
}



void fill_expansion_structures(tree *t, expansion_data *e)
{
	move moves[MAX_MOVES];
	UINT_64 hashes[MAX_MOVES];
	float weights[MAX_MOVES];
	board b,c;
	int moves_num;
	int i,wins;


	bytes_to_board(e->b, b);

	e->value = get_board_value(b, &past_episodes);
	e->walk_id = -1;

	moves_num = find_possible_moves(b, moves, &(e->has_corral));

	e->moves_num = moves_num;

	get_hashes_of_moves(b, moves_num, moves, hashes);
	get_weights_of_moves(b, moves, moves_num, weights, &past_episodes);

	e->move_hash_place = t->move_hashes_num;
	for (i = 0; i < moves_num; i++)
	{
		t->move_hashes[t->move_hashes_num + i].move = moves[i];
		t->move_hashes[t->move_hashes_num + i].hash = hashes[i];
		t->move_hashes[t->move_hashes_num + i].weight = weights[i];
		t->move_hashes[t->move_hashes_num + i].deadlocked = 0;

		if (find_node_by_hash(t, hashes[i]) == -1) // new node in tree
		{
			bytes_to_board(e->b, c);
			apply_move(c, moves + i);
			wins = board_is_solved(c);
			add_node_to_tree(t, hashes[i], wins);
		}
	}
	t->move_hashes_num += moves_num;

	if (t->move_hashes_num > t->max_move_hashes)
		exit_with_error("max move-hashes reached");
}


void set_root(tree *t, board b)
{
	UINT_64 hash;
	expansion_data *e;
	int node_place;
	int nodes_before_expansion;
	int wins;

	nodes_before_expansion = t->nodes_num;

	hash = get_board_hash(b);
	wins = board_is_solved(b);

	node_place = find_node_by_hash(t, hash);
	if (node_place == -1)
	{
		add_node_to_tree(t, hash, wins);
	}
	else
	{
		if (t->nodes[node_place].expansion != -1)
			exit_with_error("root is already in tree");
	}

	node_place = find_node_by_hash(t, hash);

	t->nodes[node_place].expansion = t->expansions_num;
	e = t->expansions + t->expansions_num;
	t->expansions_num++;

	e->node = t->nodes + node_place;
	e->b = t->boards + height*width*(t->expansions_num - 1);
	board_to_bytes(b, e->b);
	e->depth = 0;
	e->father = NULL;
	e->move_from_father = -1;

	e->weight = 0;
	e->best_weight = 0;

	fill_expansion_structures(t, e);

	set_deadlock_status(t, e);
}


void expand_node(tree *t, expansion_data *e, int son)
{
	board b;
	move_hash_data *mh;
	UINT_64 hash;
	int node_place;
	node_element *n;
	expansion_data *next;
	int nodes_before_expansion;
	move move_to_play;

	double move_weight;

	move_weight = get_weight_of_unexpanded_son(t, e, son);

	nodes_before_expansion = t->nodes_num;

	mh = t->move_hashes + e->move_hash_place;
	hash = mh[son].hash;

	node_place = find_node_by_hash(t, hash);
	if (node_place == -1) exit_with_error("missing node");
	n = t->nodes + node_place;

	if (n->expansion != -1) exit_with_error("already expanded");
	if (n->deadlocked) exit_with_error("expanding a leaf which is known to be deadlocked");

	n->expansion = t->expansions_num;
	next = t->expansions + t->expansions_num;
	t->expansions_num++;

	if (t->expansions_num >= t->max_expansions)
		exit_with_error("max expansions reached");

	bytes_to_board(e->b, b);
	move_to_play = mh[son].move;
	apply_move(b, &move_to_play);
	if (get_board_hash(b) != mh[son].hash) exit_with_error("hash error");

	next->node = n;
	next->b = t->boards + height*width*(t->expansions_num - 1);
	board_to_bytes(b, next->b);
	next->depth  = e->depth + 1;

	next->father = e;
	next->move_from_father = son;

	next->weight = e->weight + move_weight;
	next->best_weight = 0;

	fill_expansion_structures(t, next);

	set_deadlock_status(t, next);
}


void print_node(tree *t, expansion_data *e, int print_moves)
{
	int moves_num;
	int i;
	move_hash_data *mh;
	node_element *node;
	int node_place;
	board b;

	printf("Node id: %d\n", e - t->expansions);

	bytes_to_board(e->b, b);
	print_board(b);

	if (verbose < 4) return;


	printf("depth: %d\n", e->depth);
	printf("value= %f\n", e->value);

	if (e->has_corral)
		printf("node has corral\n");

	if (e->node->deadlocked)
		if (board_contains_boxes(b))
			printf("#position is deadlocked.#\n");

	if (print_moves)
	{
		mh = t->move_hashes + e->move_hash_place;

		moves_num = e->moves_num;

		printf("Node has %d moves\n", moves_num);

		for (i = 0; i < moves_num; i++)
		{
			printf("%3d: ", i);
			print_move(&(mh[i].move));

			node_place = find_node_by_hash(t, mh[i].hash);
			if (node_place == -1) exit_with_error("missing node");
			node = t->nodes + node_place;

			if (node->expansion != -1)
				printf("+");
			else
				printf("-");

			if ((mh[i].deadlocked) || (node->deadlocked))
			{
				printf("#deadlock#\n");
				continue;
			}

			printf(" weight= %lf\n", mh[i].weight);
		}
	}
}


move get_move_to_node(tree *t, expansion_data *e)
{
	expansion_data *f;
	move_hash_data *mh;
	int i;

	if (e->father == 0) exit_with_error("no father");
	f = e->father;
	mh = t->move_hashes + f->move_hash_place;

	i = e->move_from_father;
	return mh[i].move;
}


expansion_data *last_expansion(tree *t)
{
	return t->expansions + t->expansions_num - 1;
}

int has_a_winning_son(tree *t, expansion_data *e, int *son)
{
	int i;
	move_hash_data *mh = t->move_hashes + e->move_hash_place;
	node_element *n;
	int place;

	*son = -1;

	for (i = 0; i < e->moves_num; i++)
	{
		if (mh[i].deadlocked) continue;

		place = find_node_by_hash(t, mh[i].hash);
		if (place == -1) exit_with_error("missing node");
		n = t->nodes + place;

		if (n->wins)
		{
			*son = i;
			return 1;
		}		

	}
	return 0;
}

node_element *get_son(tree *t, expansion_data *e, int son)
{
	move_hash_data *mh;
	int place;

	mh = t->move_hashes + e->move_hash_place;
	place = find_node_by_hash(t, mh[son].hash);
	if (place == -1) exit_with_error("missing node");
	return t->nodes + place;
}



int son_is_expanded(tree *t, expansion_data *e, int son)
{
	move_hash_data *mh;
	int place;
	node_element *n;

	mh = t->move_hashes + e->move_hash_place;
	place = find_node_by_hash(t, mh[son].hash);
	if (place == -1) exit_with_error("missing son");
	n = t->nodes + place;

	return (n->expansion == -1 ? 0 : 1);
}


int son_is_deadlocked(tree *t, expansion_data *e, int son)
{
	move_hash_data *mh;
	int place;
	node_element *n;

	mh = t->move_hashes + e->move_hash_place;
	place = find_node_by_hash(t, mh[son].hash);
	if (place == -1) exit_with_error("missing son");
	n = t->nodes + place;

	return n->deadlocked;
}



expansion_data* go_to_son(tree *t, expansion_data *e, int son)
{
	move_hash_data *mh;
	int place;
	node_element *n;

	mh = t->move_hashes + e->move_hash_place;
	place = find_node_by_hash(t, mh[son].hash);
	if (place == -1) exit_with_error("missing son");
	n = t->nodes + place;

	if (n->deadlocked) exit_with_error("going to deadlocked son");
	if (n->expansion == -1) exit_with_error("going to unexpanded son");

	return t->expansions + n->expansion;
}
